/**
 * All this is mostly based on rlyeh's minimal library.
 * Copied here to review all his code and understand what's going on.
**/

/*

  GP2X minimal library v0.A by rlyeh, (c) 2005. emulnation.info@rlyeh (swap it!)

  Thanks to Squidge, Robster, snaff, Reesy and NK, for the help & previous work! :-)

  License
  =======

  Free for non-commercial projects (it would be nice receiving a mail from you).
  Other cases, ask me first.

  GamePark Holdings is not allowed to use this library and/or use parts from it.

*/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <errno.h>

#include "gp2x.h"
#include "usbjoy.h"
#include "squidgehack.h"

static volatile unsigned short *gp2x_memregs;
static volatile unsigned long  *gp2x_memregl;
static volatile unsigned long  *gp2x_blitter;
static void *gp2x_screens[4];
static int screensel = 0;
static int memdev = 0;
// static int sounddev = 0, mixerdev = 0; // ?

void *gp2x_screen;

#define FRAMEBUFF_WHOLESIZE (0x20000*5) // 320*240 + some more
#define FRAMEBUFF_ADDR0 (0x4000000-FRAMEBUFF_WHOLESIZE)
#define FRAMEBUFF_ADDR1 (FRAMEBUFF_ADDR0+0x20000)
#define FRAMEBUFF_ADDR2 (FRAMEBUFF_ADDR1+0x20000)
#define FRAMEBUFF_ADDR3 (FRAMEBUFF_ADDR2+0x20000)

static const int gp2x_screenaddrs[] = { FRAMEBUFF_ADDR0, FRAMEBUFF_ADDR1, FRAMEBUFF_ADDR2, FRAMEBUFF_ADDR3 };

// old reg vals
static unsigned short gp2x_screenaddr_old[4];
static unsigned short oldval_90a;
static unsigned short oldval_904;

/* utils */
extern void spend_cycles(int c);


/* video stuff */
void gp2x_video_flip(void)
{
	unsigned short msw16 = (gp2x_screenaddrs[screensel&3] >> 16);

#ifdef USE_BLITTER
	// make sure blitter is done
	while (gp2x_blitter[0x0034 >> 2] & 1)
		spend_cycles(512);
#endif

	gp2x_memregs[0x2910>>1] = msw16;
	gp2x_memregs[0x2914>>1] = msw16;

	// jump to other buffer:
	gp2x_screen = gp2x_screens[++screensel&3];
}


void gp2x_video_changemode(int bpp)
{
  	gp2x_memregs[0x28DA>>1]=(((bpp+1)/8)<<9)|0xAB; /*8/15/16/24bpp...*/
  	gp2x_memregs[0x290C>>1]=320*((bpp+1)/8); /*line width in bytes*/

  	gp2x_memset_all_buffers(0, 0, 320*240);
	gp2x_video_flip();
}


void gp2x_video_setpalette(int *pal, int len)
{
	unsigned short *g=(unsigned short *)pal;
	volatile unsigned short *memreg = &gp2x_memregs[0x295A>>1];
	gp2x_memregs[0x2958>>1] = 0;

	len *= 2;
	while(len--) *memreg=*g++;
}


// TV Compatible function //
void gp2x_video_RGB_setscaling(int W, int H)
{
	float escalaw, escalah;
	int bpp = (gp2x_memregs[0x28DA>>1]>>9)&0x3;

	escalaw = 1024.0; // RGB Horiz LCD
	escalah = 320.0; // RGB Vert LCD

	if(gp2x_memregs[0x2800>>1]&0x100) //TV-Out
	{
		escalaw=489.0; // RGB Horiz TV (PAL, NTSC)
		if (gp2x_memregs[0x2818>>1]  == 287) //PAL
			escalah=274.0; // RGB Vert TV PAL
		else if (gp2x_memregs[0x2818>>1]  == 239) //NTSC
			escalah=331.0; // RGB Vert TV NTSC
	}

	// scale horizontal
	gp2x_memregs[0x2906>>1]=(unsigned short)((float)escalaw *(W/320.0));
	// scale vertical
	gp2x_memregl[0x2908>>2]=(unsigned long)((float)escalah *bpp *(H/240.0));
}


void gp2x_memset_all_buffers(int offset, int byte, int len)
{
	memset((char *)gp2x_screens[0] + offset, byte, len);
	memset((char *)gp2x_screens[1] + offset, byte, len);
	memset((char *)gp2x_screens[2] + offset, byte, len);
	memset((char *)gp2x_screens[3] + offset, byte, len);
}


unsigned long gp2x_joystick_read(void)
{
	int i;
  	unsigned long value=(gp2x_memregs[0x1198>>1] & 0x00FF);
  	if(value==0xFD) value=0xFA;
  	if(value==0xF7) value=0xEB;
  	if(value==0xDF) value=0xAF;
  	if(value==0x7F) value=0xBE;
  	value = ~((gp2x_memregs[0x1184>>1] & 0xFF00) | value | (gp2x_memregs[0x1186>>1] << 16));

	if (num_of_joys > 0) {
		// check the usb joy as well..
		gp2x_usbjoy_update();
		for (i = 0; i < num_of_joys; i++)
			value |= gp2x_usbjoy_check(i);
	}

	return value;
}


static void gp2x_dummy_blit(void)
{
	// Blitter seems to have a 16 byte buffer, so force flushing by
	// drawing 8 16bit pixels onto themselves.
	gp2x_blitter[0x0000>>2] = (1 << 5) | (1 << 6);
	gp2x_blitter[0x0004>>2] = 0x3101000; // front screen buffer
	gp2x_blitter[0x000C>>2] = 1<<8;
	gp2x_blitter[0x0020>>2] = 0;
	gp2x_blitter[0x002C>>2] = (1 << 16) | 8; // number of bytes to copy
	gp2x_blitter[0x0030>>2] = 1<< 8 | 1<< 9 | 0xaa;
	// Make sure blitter is ready
	while (gp2x_blitter[0x0034 >> 2] & 1)
	{
		asm volatile ("nop");
		asm volatile ("nop");
		asm volatile ("nop");
		asm volatile ("nop");
	}
	asm volatile ("" ::: "memory");
	gp2x_blitter[0x0034 >> 2] = 0x0001;
}


extern unsigned char *buf;
extern unsigned char *lpbuf, *rpbuf;

void noiz2sa_blit(void)
{
	unsigned int n2buffer_base = gp2x_screenaddrs[3]+0x20000;
	unsigned int screen_base = gp2x_screenaddrs[screensel&3];

	// make sure blitter is not busy
	while (gp2x_blitter[0x0034 >> 2] & 1)
		spend_cycles(512);

	gp2x_blitter[0x0000>>2] = 0;				// no ROP, 8bpp, no fraction for dst
	gp2x_blitter[0x0004>>2] = screen_base;			// dst is current framebuffer
	gp2x_blitter[0x0008>>2] = 320;				// dst stride
	gp2x_blitter[0x000C>>2] = (1 << 8) | (1 << 7);		// src is framebuffer, use ROP
	gp2x_blitter[0x0010>>2] = n2buffer_base + 160*240;	// src is lpbuffer
	gp2x_blitter[0x0014>>2] = 80;				// src stride
	gp2x_blitter[0x0020>>2] = 0;				// no patterns
	gp2x_blitter[0x002C>>2] = (240 << 16) | 80;		// lpbuffer is 80x240
	gp2x_blitter[0x0030>>2] = (1 << 8) | (1 << 9) | 0xCC;	// incrementing addresses, copy ROP
	gp2x_blitter[0x0034>>2] = 1;				// go!

	// wait for it to finish
	while (gp2x_blitter[0x0034 >> 2] & 1)
		spend_cycles(512);

	gp2x_blitter[0x0004>>2] = screen_base + 80 + 160;
	gp2x_blitter[0x0010>>2] = n2buffer_base + 240*240;	// src is rpbuffer
	gp2x_blitter[0x0034>>2] = 1;				// go!

	// wait for it to finish
	while (gp2x_blitter[0x0034 >> 2] & 1)
		spend_cycles(512);

	gp2x_blitter[0x0004>>2] = screen_base + 80;
	gp2x_blitter[0x0010>>2] = n2buffer_base;		// src is buf
	gp2x_blitter[0x0014>>2] = 160;				// buf stride
	gp2x_blitter[0x002C>>2] = (240 << 16) | 160;		// buf is 160x240
	gp2x_blitter[0x0034>>2] = 1;				// go!

	// flip will wait for last blit to finish
}


static void noiz2sa_fill_panel(unsigned int dst_addr)
{
	// make sure blitter is not busy
	while (gp2x_blitter[0x0034 >> 2] & 1)
		spend_cycles(512);

	gp2x_blitter[0x0000>>2] = 0;				// no ROP, 8bpp, no fraction for dst
	gp2x_blitter[0x0004>>2] = dst_addr;			// dst
	gp2x_blitter[0x0008>>2] = 80;				// dst stride
	gp2x_blitter[0x000C>>2] = 0;				// no src
	gp2x_blitter[0x0020>>2] = 0;				// no patterns
	gp2x_blitter[0x002C>>2] = (240 << 16) | 80;		// lpbuffer is 80x240
	gp2x_blitter[0x0030>>2] = (1 << 8) | (1 << 9) | 0;	// incrementing addresses, fill ROP
	gp2x_blitter[0x0034>>2] = 1;				// go!

	while (gp2x_blitter[0x0034 >> 2] & 1)
		spend_cycles(512);
}

void noiz2sa_fill_lp(void)
{
	noiz2sa_fill_panel(gp2x_screenaddrs[3] + 0x20000 + 160*240);
}

void noiz2sa_fill_rp(void)
{
	noiz2sa_fill_panel(gp2x_screenaddrs[3] + 0x20000 + 240*240);
}

/* common */
void gp2x_init(void)
{
	printf("entering init()\n"); fflush(stdout);

  	memdev = open("/dev/mem", O_RDWR);
	if (memdev == -1)
	{
		printf("open(\"/dev/mem\") failed with %i\n", errno);
		exit(1);
	}

	gp2x_memregs = mmap(0, 0x10000, PROT_READ|PROT_WRITE, MAP_SHARED, memdev, 0xc0000000);
	printf("memregs are @ %p\n", gp2x_memregs);
	if(gp2x_memregs == MAP_FAILED)
	{
		printf("mmap(memregs) failed with %i\n", errno);
		exit(1);
	}
	gp2x_memregl = (unsigned long *) gp2x_memregs;

  	gp2x_screens[0] = mmap(0, FRAMEBUFF_WHOLESIZE, PROT_WRITE, MAP_SHARED, memdev, FRAMEBUFF_ADDR0);
	if(gp2x_screens[0] == MAP_FAILED)
	{
		printf("mmap(gp2x_screen) failed with %i\n", errno);
		exit(1);
	}
	printf("framebuffers point to %p\n", gp2x_screens[0]);
	gp2x_screens[1] = (char *) gp2x_screens[0]+0x20000;
	gp2x_screens[2] = (char *) gp2x_screens[1]+0x20000;
	gp2x_screens[3] = (char *) gp2x_screens[2]+0x20000;

	gp2x_screen = gp2x_screens[0];
	screensel = 0;

	mmuhack();

	// for noiz2sa
	buf   = gp2x_screens[3]+0x20000;
	lpbuf = buf   + 160*240;
	rpbuf = lpbuf + 80*240;

	gp2x_screenaddr_old[0] = gp2x_memregs[0x290E>>1];
	gp2x_screenaddr_old[1] = gp2x_memregs[0x2910>>1];
	gp2x_screenaddr_old[2] = gp2x_memregs[0x2912>>1];
	gp2x_screenaddr_old[3] = gp2x_memregs[0x2914>>1];

	// always 0 because addresses are aligned
	gp2x_memregs[0x290E>>1] = gp2x_memregs[0x2912>>1] = 0;

	gp2x_blitter = mmap(0, 0x100, PROT_READ|PROT_WRITE, MAP_SHARED, memdev, 0xE0020000); // 0xE0024000 ??
	if(gp2x_blitter == MAP_FAILED)
	{
		printf("mmap(blitter) failed with %i\n", errno);
		exit(1);
	}

	oldval_90a = gp2x_memregs[0x090a >> 1];
	oldval_904 = gp2x_memregs[0x0904 >> 1];

	gp2x_video_RGB_setscaling(320, 240);

	// enable all video and graphic devices
	gp2x_memregs[0x090a >> 1] = 0xFFFF;
	// enable fastio
	gp2x_memregs[0x0904 >> 1] |= (1 << 10);

	printf("flushing blitter's cache..\n");
	gp2x_dummy_blit();

	gp2x_usbjoy_init();

	printf("exitting init()\n"); fflush(stdout);
}


void gp2x_deinit(void)
{
	printf("gp2x_deinit()\n");
	gp2x_video_changemode(15);

	gp2x_memregs[0x290E>>1] = gp2x_screenaddr_old[0];
	gp2x_memregs[0x2910>>1] = gp2x_screenaddr_old[1];
	gp2x_memregs[0x2912>>1] = gp2x_screenaddr_old[2];
	gp2x_memregs[0x2914>>1] = gp2x_screenaddr_old[3];
	gp2x_memregs[0x090a>>1] = oldval_90a;
	gp2x_memregs[0x0904>>1] = oldval_904;

	mmuunhack();

	munmap(gp2x_screens[0], FRAMEBUFF_WHOLESIZE);
	munmap((void *)gp2x_memregs, 0x10000);
	munmap((void *)gp2x_blitter, 0x100);
	close(memdev);

	gp2x_usbjoy_deinit();

	printf("all done, running gp2xmenu\n");

	chdir("/usr/gp2x");
	execl("gp2xmenu", "gp2xmenu", NULL);
}



