/*
 * $Id: noiz2sa.c,v 1.8 2003/02/12 13:55:13 kenta Exp $
 *
 * Copyright 2002 Kenta Cho. All rights reserved.
 */

/**
 * Noiz2sa main routine.
 *
 * @version $Revision: 1.8 $
 */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/time.h>

#include "noiz2sa.h"
#include "screen.h"
#include "vector.h"
#include "ship.h"
#include "shot.h"
#include "frag.h"
#include "bonus.h"
#include "foe_mtd.h"
#include "brgmng_mtd.h"
#include "background.h"
#include "degutil.h"
#include "soundmanager.h"
#include "attractmanager.h"

#include "gp2x/gp2x.h"

static int noSound = 0;
static int showFps = 0;

// Initialize and load preference.
static void initFirst() {
  struct timeval tval;

  loadPreference();
  gettimeofday(&tval, 0);
  srand(tval.tv_usec*7/8);
  initBarragemanager();
  initAttractManager();
}

// Quit and save preference.
void quitLast() {
  if ( !noSound ) closeSound();
  savePreference();
  closeBarragemanager();
  gp2x_deinit();
  exit(0);
}

int status;

static float stagePrm[STAGE_NUM+ENDLESS_STAGE_NUM+1][3] = {
  {13, 0.5f, 0.12f}, {2, 1.8f, 0.15f}, {3, 3.2f, 0.1f}, {90, 6.0f, 0.3f}, {5, 5.0f, 0.6f},
  {6, 10.0f, 0.6f}, {7, 5.0f, 2.2f}, {98, 12.0f, 1.5f}, {9, 10.0f, 2.0f}, {79, 21.0f, 1.5f},
  {-3, 5.0f, 0.7f}, {-1, 10.0f, 1.2f}, {-4, 15.0f, 1.8f}, {-2, 16.0f, 1.8f},
  {0, -1.0f, 0.0f},
};

void initTitleStage(int stg) {
  initFoes();
  initBarrages(stagePrm[stg][0], stagePrm[stg][1], stagePrm[stg][2]);
}

void initTitle() {
  int stg;
  status = TITLE;

  stg = initTitleAtr();
  initShip();
  initShots();
  initFrags();
  initBonuses();
  initBackground();
  setStageBackground(1);

  initTitleStage(stg);
  stopMusic();
  savePreference();
  playMusic(7);
}

void initGame(int stg) {
  status = IN_GAME;

  initShip();
  initShots();
  initFoes();
  initFrags();
  initBonuses();
  initBackground();

  initBarrages(stagePrm[stg][0], stagePrm[stg][1], stagePrm[stg][2]);
  initGameState(stg);
  if ( stg < STAGE_NUM ) {
    setStageBackground(stg%5+1);
    playMusic(stg%5+1);
  } else {
    if ( !insane ) {
      setStageBackground(0);
      playMusic(0);
    } else {
      setStageBackground(6);
      playMusic(6);
    }
  }
}

void initGameover() {
  status = GAMEOVER;
  initGameoverAtr();
}

void initStageClear() {
  status = STAGE_CLEAR;
  initStageClearAtr();
}

static void move() {
  switch ( status ) {
  case TITLE:
    moveTitleMenu();
    moveBackground();
    addBullets();
    moveFoes();
    break;
  case IN_GAME:
    moveBackground();
    addBullets();
    moveShots();
    moveShip();
    moveFoes();
    moveFrags();
    moveBonuses();
    break;
  case GAMEOVER:
    moveGameover();
    moveBackground();
    addBullets();
    moveShots();
    moveFoes();
    moveFrags();
    break;
  case STAGE_CLEAR:
    moveStageClear();
    moveBackground();
    moveShots();
    moveShip();
    moveFrags();
    moveBonuses();
    break;
  case PAUSE:
    movePause();
    break;
  }
}

static void draw() {
  switch ( status ) {
  case TITLE:
    // Draw background.
    drawBackground();
    drawFoes();
    drawBulletsWake();
    blendScreen();
    // Draw forground.
    drawBullets();
    drawScore();
    drawTitleMenu();
    break;
  case IN_GAME:
    // Draw background.
    drawBackground();
    drawBonuses();
    drawFoes();
    drawBulletsWake();
    drawFrags();
    blendScreen();
    // Draw forground.
    drawShots();
    drawShip();
    drawBullets();
    drawScore();
    break;
  case GAMEOVER:
    // Draw background.
    drawBackground();
    drawFoes();
    drawBulletsWake();
    drawFrags();
    blendScreen();
    // Draw forground.
    drawShots();
    drawBullets();
    drawScore();
    drawGameover();
    break;
  case STAGE_CLEAR:
    // Draw background.
    drawBackground();
    drawBonuses();
    drawFrags();
    blendScreen();
    // Draw forground.
    drawShots();
    drawShip();
    drawScore();
    drawStageClear();
    break;
  case PAUSE:
    // Draw background.
    drawBackground();
    drawBonuses();
    drawFoes();
    drawBulletsWake();
    drawFrags();
    blendScreen();
    // Draw forground.
    drawShots();
    drawShip();
    drawBullets();
    drawScore();
    drawPause();
    break;
  }
  if (showFps) drawFps();
}

static void usage(char *argv0) {
  fprintf(stderr, "Usage: %s [-nosound] [-reverse] [-brightness n] [-nowait] [-insanespeed]\n"
    "  [-showfps] [-musicvolume n]\n", argv0);
}

static void parseArgs(int argc, char *argv[]) {
  int i;
  for ( i=1 ; i<argc ; i++ ) {
    if ( strcmp(argv[i], "-nosound") == 0 ) {
      noSound = 1;
    } else if ( strcmp(argv[i], "-window") == 0 ) {
      // ignore
    } else if ( strcmp(argv[i], "-reverse") == 0 ) {
      buttonReversed = 1;
    } else if ( (strcmp(argv[i], "-brightness") == 0) && argv[i+1] ) {
      i++;
      brightness = (int)atoi(argv[i]);
      if ( brightness < 0 || brightness > 256 ) {
	brightness = DEFAULT_BRIGHTNESS;
      }
    } else if ( strcmp(argv[i], "-nowait") == 0 ) {
      nowait = 1;
    } else if ( strcmp(argv[i], "-accframe") == 0 ) {
      // ignore
    } else if ( strcmp(argv[i], "-insanespeed") == 0 ) {
      insanespeed = 1;
    } else if ( strcmp(argv[i], "-showfps") == 0 ) {
      showFps = 1;
    } else if ( strcmp(argv[i], "-musicvolume") == 0 ) {
      i++;
      music_volume = atoi(argv[i]);
    } else {
      usage(argv[0]);
      exit(1);
    }
  }
}


unsigned long keys;

static int updateKeys(void)
{
  static int pPrsd = 0;
  int done = 0;

  keys = gp2x_joystick_read();

  if ((keys & (GP2X_START|GP2X_PUSH)) == (GP2X_START|GP2X_PUSH) ||
      (keys & (GP2X_L|GP2X_R|GP2X_START)) == (GP2X_L|GP2X_R|GP2X_START))
        return 1; // done

  if (keys & GP2X_VOL_UP) increaseSoundVolume();
  else if (keys & GP2X_VOL_DOWN) decreaseSoundVolume();

  if (keys & GP2X_START) {
    if ( !pPrsd ) {
      if ( status == IN_GAME ) {
        pauseMusic();
        status = PAUSE;
      } else if ( status == PAUSE ) {
        unpauseMusic();
        status = IN_GAME;
      }
    }
    pPrsd = 1;
  } else {
    pPrsd = 0;
  }

  return done;
}


int interval = INTERVAL_BASE;
int tick = 0;
int fps = 0;
char **g_argv;
void simple_wait(int thissec, int lim_time);

int main(int argc, char *argv[]) {
  int done = 0;
  int thissec = 0;
  int frames_done = 0, frames_shown = 0;
  int lim_time = 0;
  struct timeval tval; // timing

  g_argv = argv;

  parseArgs(argc, argv);
  initDegutil();
  gp2x_init();
  gp2x_video_changemode(8);
  if ( !noSound ) {
	  initSound();
  }
  init_screen();
  initFirst();
  initTitle();


  while ( !done )
  {
    gettimeofday(&tval, 0);

    // second changed?
    if(thissec != tval.tv_sec) {
      thissec = tval.tv_sec;
      fps = frames_shown;
      frames_done = frames_shown = 0;
      lim_time = 0;
    }

    lim_time += 1000000/60+1;
    lim_time += (interval - INTERVAL_BASE) * 970;
    if(tval.tv_usec > lim_time) {
      // no time left for this frame - skip
      done = updateKeys();
      move();
      tick++;
      frames_done++;
      continue;
    }

    done = updateKeys();
    move();
    tick++;
    smokeScreen();
    draw();
    prepareScreen();
    updateSound();

    // sleep if we are still too fast
    // usleep sleeps for ~20ms minimum, so it is not a solution here
    gettimeofday(&tval, 0);
    if(thissec != tval.tv_sec) tval.tv_usec+=1000000;
    if(tval.tv_usec < lim_time)
    {
      // we are too fast
      simple_wait(thissec, lim_time);
    }

    gp2x_video_flip();

    frames_done++; frames_shown++;
  }

  quitLast();
  return (0);
}

// used instead usleep() which doesn't appear to be accurate
void simple_wait(int thissec, int lim_time)
{
  extern void spend_cycles(int c);
  struct timeval tval;

  spend_cycles(1024);
  gettimeofday(&tval, 0);
  if (thissec != tval.tv_sec)
    tval.tv_usec += 1000000;

  while (tval.tv_usec < lim_time)
  {
    spend_cycles(1024);
    gettimeofday(&tval, 0);
    if (thissec != tval.tv_sec)
      tval.tv_usec += 1000000;
  }
}

